/*
 * ported to Linux from Francis Dupont (INRIA Rocquencourt)'s IPv6 
 * implementation on Net/FreeBSD.
 *
 * Last updates:
 *
 * 08/09
 */

#include "ndpd.h"

struct timeval now;             /* Time struct             */
int s, r, u;			/* Socket file descriptors */
int pid;                        /* Application pid         */

int verbose = 0;                /* Verbose mode */
int debug = 0;                  /* Debug mode   */

char usage[] = "Usage: ndpd-host [-d] [-v]\n";
char inet_ntop_ret[INET6_ADDRSTRLEN];
char inet_ntop_ret1[INET6_ADDRSTRLEN];

void main(int argc, char *argv[]) {
   char ch;
   u_int val;
   int nfds, 
      cc,
      offset;
   fd_set rfds;
   struct timeval tv;
   struct timeout *tp;
   struct interface *ifp;
   extern int optind;

   pid = getpid();
   (void) setpriority(PRIO_PROCESS, 0, -10);
   (void) gettimeofday(&now, NULL);

   /*
    * Get input parameters
    */
   while ((ch = getopt(argc, argv, "dv")) != EOF)
      switch (ch) {
      case 'd':
	 debug = 1;
	 break;
      case 'v':
	 verbose++;
	 break;
      default:
	 (void) fprintf(stderr, usage);
	 exit(1);
      }

   argc -= optind;
   argv += optind;

   if (argc != 0) {
      (void) fprintf(stderr, "Extra parameters\n");
      (void) fprintf(stderr, usage);
      exit(1);
   }

   srandom((int)gethostid());
   setlinebuf(stdout);

   /*
    * Open communication socket
    */
   if ((s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
      logperror("socket (ICMP)");
      exit(2);
   }
   
   /* Include header both ways */
   val = 2;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_HDRINCL,
		  &val, sizeof(val)) < 0) {
      logperror("setsockopt (IPV6_HDRINCL)");
      exit(3);
   }

   /* Suppress csum computation by kernel */
   offset = -1;            
   if (setsockopt(s, IPPROTO_IPV6, IPV6_CHECKSUM, 
                       &offset, sizeof(offset)) < 0) {
      logperror("setsockopt (IPV6_CHECKSUM)");
      exit(3);
   }
   
   /* Receive interface info */
   val = 1;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_PKTINFO,
		  &val, sizeof(val)) < 0) {
      logperror("setsockopt (IPV6_RXINFO)");
      exit(3);
   }
	 
   /* Set hop number */
   val = ICMP6_ND_HOPS;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		  &val, sizeof(val)) < 0) {
      logperror("setsockopt (IPV6_MULTICAST_HOPS)");
      exit(3);
   }
       
   /* Loopback multicast messages */
   val = 1;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
		  &val, sizeof(val)) < 0) {
      logperror("setsockopt (IPV6_MULTICAST_LOOP)");
      exit(3);
   }
                
   /* 
    * Open ioctl socket
    */
   if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
      logperror("socket (UDP)");
      exit(2);
   }
   (void) shutdown(u, 0);

   /*
    * Open communication link with kernel
    */
   if((r = open("/dev/route", O_RDWR)) < 0) {
      logperror("Cannot open netlink device\n");
      exit(3);/* netlink... choose a MINOR! */
   }
	
   /*
    * Get system configuration
    */
   init();
   dump();
   
   if (verbose)
      fprintf(stderr, "Init complete\n");
	
   if( !interfaces.tqh_first) {
      if(verbose) {
	 logerr("No IPV6 interface !!!\n");
	 logerr("You may want to run autoconf.\n");
	 exit(1);
      }
   }

   /*
    * Set signals
    */
   (void) signal(SIGUSR1, set_verbose);
   (void) signal(SIGUSR2, set_verbose);
   (void) signal(SIGINT, dump);
   (void) signal(SIGTERM, finish);
   (void) signal(SIGQUIT, finish);

   /*
    * Get main interface
    */
   get_main_if();   

   (void) gettimeofday(&now, NULL);
        
   /* 
    * Start soliciting router advertisements 
    */
   for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
      if (ifp->flags & IFF_UP)
	 send_rtsol(ifp);

   if (Calloc(tp, timeout) == NULL) {
      logerr("out of memory\n");
      abort();
   }
   (void) gettimeofday(&now, NULL);
   tp->tv.tv_sec = now.tv_sec + WAIT_INIT;
   tp->tv.tv_usec = now.tv_usec;
   tp->func = reset_default;
   add_timeout(tp);

   /*
    * Main loop
    */
   for (;;) {
      (void) gettimeofday(&now, NULL);

      /*
       * Check timeout list
       */
      tp = timeouts.tqh_first;
      if (tp && AFTER(now, tp->tv)) {
	 TAILQ_REMOVE(&timeouts, tp, tm_list);
	 tp->expired++;
	 (*tp->func) (tp->arg);
	 if (tp->refcount <= 0)
	    free(tp);
	 continue;
      }
                
      /*
       * Post select flags
       */
      FD_ZERO(&rfds);
      FD_SET(s, &rfds);
      FD_SET(r, &rfds);
      nfds = r > s ? r : s;

      /*
       * Compute select timer
       */
      if (tp) {
	 tv.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
	 tv.tv_sec = tp->tv.tv_sec - now.tv_sec;
	 NORMTIME(tv);
	 if (tv.tv_sec > 0)
	    tv.tv_sec -= 1;
	 else
	    tv.tv_usec = 10000;
      } else {
	 tv.tv_sec = 3;
	 tv.tv_usec = 0;
      }
      
      /*
       * Select call
       */
      if (verbose)
	 fprintf(stderr, "Select...\n");
      cc = select(nfds + 1, &rfds, NULL, NULL, &tv);

      /*
       * Select result
       */
      /* error */
      if (cc < 0) {
	 if (verbose)
	    logperror("select");
	 continue;
      }
      
      /* Nothing new under the stars */
      if (cc == 0)
	 continue;
      
      /* 
       * Someone rang... 
       */
      /* ICMP message */
      if (FD_ISSET(s, &rfds))
	 recv_icmp();
      
      /* Kernel message */
      if (FD_ISSET(r, &rfds))
	 recv_route();

      if (verbose)
	 logerr("end of main loop reached\n");
   }
   /*NOTREACHED*/
}

void
set_verbose(sig)
	int sig;
{
	if (sig == SIGUSR1)
		verbose++;
	else
		verbose = 0;
}

static const char wday_name[7][3] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
static const char mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

void dump()
{
	register struct timeout *tp;
	register struct interface *ifp;
	struct tm *tmnow;

	(void) gettimeofday(&now, NULL);
	tmnow = localtime((const time_t *)&now.tv_sec);

	logdebug("DUMP at %.3s %.3s%3d %02d:%02d:%02d.%06d %d\n",
		 wday_name[tmnow->tm_wday],
		 mon_name[tmnow->tm_mon],
		 tmnow->tm_mday, tmnow->tm_hour,
		 tmnow->tm_min, tmnow->tm_sec,
		 now.tv_usec,
		 1900 + tmnow->tm_year);

	logdebug("timeout list:\n");
	for (tp = timeouts.tqh_first; tp; tp = tp->tm_list.tqe_next)
		dump_timeout(tp);

	logdebug("interface list:\n");
	for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
		dump_interface(ifp);

	if (def_if)
		logdebug("default interface is %s\n", def_if->name);
	else
		logdebug("no current default interface\n");
	if (main_if)
		logdebug("main interface is %s\n", main_if->name);
	else
		logdebug("no main interface\n");

	logdebug("default source (link#%d) = %s\n",
		 def_ifa.index,
		 inet_ntop(AF_INET6, &def_ifa.addr,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
}

void
finish()
{
	add_default(0);
	if (verbose)
		logtrace("finished!\n");
	exit(0);
}



















